import sqlalchemy
from pg8000.exceptions import DatabaseError


def deploy(db_engine: sqlalchemy.engine.base.Engine, schema: str, reset: bool = False):
    if reset:
        _kill(db_engine, schema)
    _create_schema(db_engine, schema)
    _create_types(db_engine, schema)
    _create_tables(db_engine, schema)
    _create_functions(db_engine, schema)


def _create_schema(db_engine: sqlalchemy.engine.base.Engine, schema: str):
    with db_engine.connect() as conn:
        conn.execute(sqlalchemy.schema.CreateSchema(schema, if_not_exists=True))
        conn.commit()


def _create_types(db_engine: sqlalchemy.engine.base.Engine, schema: str):
    # Arguments
    try:
        with db_engine.connect() as conn:
            conn.execute(
                sqlalchemy.text(
                    f"""
                    CREATE TYPE jobmaster_argument AS (
                        arg_key     text,
                        arg_value   json
                    );
                    """
                )
            )
            conn.commit()
    except DatabaseError as e:
        pass


def _create_tables(db_engine: sqlalchemy.engine.base.Engine, schema: str):
    with db_engine.connect() as conn:

        # Task Definitions
        conn.execute(
            sqlalchemy.text(
                f"""
                CREATE TABLE IF NOT EXISTS {schema}.task_types (
                    type_key        text NOT NULL,
                    help            text
                );
                
                CREATE TABLE IF NOT EXISTS {schema}.tasks (
                    type_key        text NOT NULL,
                    task_key        text NOT NULL,
                    help            text,
                    process_units   integer,
                    write_all       text ARRAY
                );
                
                CREATE TABLE IF NOT EXISTS {schema}.parameters (
                    type_key        text NOT NULL,
                    task_key        text NOT NULL,
                    parameter_key   text NOT NULL,
                    value_type      text,
                    select_from     json ARRAY,
                    default_value   json,
                    required        boolean DEFAULT false,
                    write_all       boolean DEFAULT false,
                    help            text DEFAULT 'No documentation available'
                );
                """
            )
        )

        # Task Dependency Definitions
        conn.execute(
            sqlalchemy.text(
                f"""
                        CREATE TABLE IF NOT EXISTS {schema}.dependencies (
                            type_key                text NOT NULL,
                            task_key                text NOT NULL,
                            dependency_type_key     text NOT NULL,
                            dependency_task_key     text NOT NULL,
                            time                    integer
                        );
                        
                        CREATE TABLE IF NOT EXISTS {schema}.dependency_args_same (
                            type_key                text NOT NULL,
                            task_key                text NOT NULL,
                            dependency_type_key     text NOT NULL,
                            dependency_task_key     text NOT NULL,
                            argument                text NOT NULL
                        );
                        
                        CREATE TABLE IF NOT EXISTS {schema}.dependency_args_specified (
                            type_key                text NOT NULL,
                            task_key                text NOT NULL,
                            dependency_type_key     text NOT NULL,
                            dependency_task_key     text NOT NULL,
                            argument                text NOT NULL,
                            arg_value               json NOT NULL
                        );
                """
            )
        )

        # Jobs queue
        conn.execute(
            sqlalchemy.text(
                f"""
                CREATE TABLE IF NOT EXISTS {schema}.jobs (
                    job_id              uuid NOT NULL,
                    collect_id          uuid,
                    status              smallint DEFAULT 0 NOT NULL,
                    system_node_name    text DEFAULT 'external',
                    system_pid          integer DEFAULT 0,
                    cmd_id              text NOT NULL,
                    updated_at          timestamp DEFAULT now(),
                    priority            integer DEFAULT 0,
                    type_key            text NOT NULL,
                    task_key            text NOT NULL,
                    message             text
                );
                """
            )
        )

        # Arguments
        conn.execute(
            sqlalchemy.text(
                f"""
                CREATE TABLE IF NOT EXISTS {schema}.arguments (
                    job_id      uuid NOT NULL,
                    arg_key     text,
                    arg_value   json
                );
                """
            )
        )

        # Collect Ids
        conn.execute(
            sqlalchemy.text(
                f"""
                CREATE TABLE IF NOT EXISTS {schema}.collect_ids (
                    collect_id          uuid NOT NULL,
                    time_added          timestamp DEFAULT now()
                );
                """
            )
        )

        conn.commit()


def _create_functions(db_engine: sqlalchemy.engine.base.Engine, schema: str):
    current_jobs_function_sql = """
    CREATE OR REPLACE FUNCTION @SCHEMA_NAME.current_jobs () 
                RETURNS TABLE (
                    job_id_out uuid,
                    type_key_out text,
                    task_key_out text,
                    status_out smallint,
                    priority_out integer,
                    created_at_out timestamp,
                    updated_at_out timestamp,
                    arguments_out jobmaster_argument ARRAY,
                    system_node_name_out text,
                    system_pid_out integer,
                    cmd_id_out text
                ) 
                language plpgsql
                AS $$
                BEGIN
                    RETURN QUERY
                        WITH most_recent as (
                            SELECT      job_id, max(updated_at) as updated_at, min(updated_at) as created_at
                            FROM        @SCHEMA_NAME.jobs
                            GROUP BY    job_id
                        ), all_jobs AS (
                            SELECT  most_recent.job_id,
                                    j1.type_key,
                                    j1.task_key,
                                    j1.status,
                                    j1.priority,
                                    j1.system_node_name,
                                    j1.system_pid,
                                    j1.cmd_id,
                                    most_recent.created_at,
                                    most_recent.updated_at
                            FROM    @SCHEMA_NAME.jobs j1
                            RIGHT JOIN most_recent
                            ON  most_recent.job_id = j1.job_id and most_recent.updated_at = j1.updated_at
                        ), 
                        args1 AS (
                            SELECT  all_jobs.job_id,
                                    array_agg(row(args_table.arg_key, args_table.arg_value)::jobmaster_argument) as args
                            FROM  all_jobs
                            LEFT JOIN @SCHEMA_NAME.arguments args_table
                            ON all_jobs.job_id = args_table.job_id
                            GROUP BY all_jobs.job_id
                        )
    
                        SELECT  all_jobs.job_id,
                                all_jobs.type_key,
                                all_jobs.task_key,
                                all_jobs.status,
                                all_jobs.priority,
                                all_jobs.created_at,
                                all_jobs.updated_at,
                                args1.args,
                                all_jobs.system_node_name,
                                all_jobs.system_pid,
                                all_jobs.cmd_id
                        FROM    all_jobs
                        LEFT JOIN args1
                        ON      all_jobs.job_id = args1.job_id
                        ;
                END;
                $$;
    """.replace("@SCHEMA_NAME", schema)
    insert_procedure_sql = """
create or replace procedure @SCHEMA_NAME.insert(
    type_key_in text,
    task_key_in text,
    arguments_in json,
    priority_in integer default 1,
    system_name_in text default 'external',
    new_job_id inout uuid default null,
    message_out inout text default null
)
language plpgsql    
as $$
declare 
    conflicting_jobs uuid[];
    missing_arguments text[];
begin
    -- if no job_id was provided, generate one
    select coalesce(new_job_id, gen_random_uuid()) into new_job_id;
    
    -- check if the job is already in the queue
    if new_job_id in (select distinct job_id from @SCHEMA_NAME.jobs)
    then 
        select 'job already in queue' into message_out;
        return;
    end if;
    
    -- check the task exists
    if (select count(*) from @SCHEMA_NAME.tasks where type_key = type_key_in and task_key = task_key_in) = 0
    then
        select 'Invalid task: ' || type_key_in || '.' || task_key_in into message_out;
        select null into new_job_id;
        return;
    end if;
    
    -- check that the user provided the correct arguments
    with 
    -- create argsin table with the input data
    argsin as (
        select key as arg_key, value as arg_value, TRUE as yes from json_each(arguments_in)
    ), 
    -- select parameter_key from tasks matching the type_key and task_key
    t1 as (
        select parameter_key
        from @SCHEMA_NAME.parameters p
        where p.type_key = type_key_in and p.task_key = task_key_in and p.required = TRUE
    ), kvpairs as (
        select  t1.parameter_key,
                argsin.yes
        from t1 
        left join argsin
        on t1.parameter_key = argsin.arg_key
    )
    select array_agg(parameter_key)
    from kvpairs
    where yes is null
    into missing_arguments;
    
    if array_length(missing_arguments, 1) > 0
    then 
        select 'Missing arguments: ' || array_to_string(missing_arguments, ', ') into message_out;
        select null into new_job_id;
        return;
    end if;
    
    lock table @SCHEMA_NAME.jobs in ROW EXCLUSIVE mode;
    
    -- check if a job with the same args is already waiting/running
    with 
    -- create argsin and taskin: tables with the input data
    argsin as (
        select key as arg_key, value as arg_value from json_each(arguments_in)
    ), taskin as (
        select  type_key_in as type_key, task_key_in as task_key
    ), 
    -- get all jobs with current status 1 or 2
    wr_jobs as (
        select j.job_id, j.type_key, j.task_key 
        from @SCHEMA_NAME.jobs j
        right join (select job_id, max(updated_at) as updated_at from @SCHEMA_NAME.jobs group by job_id) k
        on j.job_id = k.job_id and j.updated_at = k.updated_at
        where j.status in (1, 2)
    ), 
    -- get jobs with the same type_key and task_key as the input
    task_confilicts as (
        select  wr_jobs.job_id
        from    taskin
        inner join wr_jobs
        on taskin.type_key = wr_jobs.type_key and taskin.task_key = wr_jobs.task_key
    ), 
    -- get the arguments of those jobs
    args_of_task_confilicts as (
        select      a.job_id, a.arg_key, a.arg_value
        from        task_confilicts
        left join   @SCHEMA_NAME.arguments a
        on task_confilicts.job_id = a.job_id
    ), 
    -- join with argsin on the key and value, then count the rows
    arg_conflicts as (
        select  a.job_id, count(*) as arg_count
        from    argsin
        inner join args_of_task_confilicts a
        on argsin.arg_key = a.arg_key and argsin.arg_value::jsonb = a.arg_value::jsonb
        group by a.job_id
    ), 
    -- if any job has the same number of matching args as there are input args, we know it has all the same args
    conflicts as (
        select a.job_id
        from   arg_conflicts a
        where  a.arg_count >= (select count(*) from argsin)
    )
    -- finally select the ids of conflicting jobs into an array
    select array_agg(job_id) from conflicts into conflicting_jobs;
    
    -- if there are any conflicting jobs, we return
    if array_length(conflicting_jobs, 1) > 0
    then 
        select 'Job with same specification already in queue. Returning job_id for that job.' into message_out;
        select conflicting_jobs[1] into new_job_id;
        return;
    end if;
    
    insert into @SCHEMA_NAME.jobs
    values (new_job_id, null, 1, system_name_in, 0, 'external', now(), priority_in, type_key_in, task_key_in, 'Inserted using insert_job procedure');
    
    insert into @SCHEMA_NAME.arguments
    select  new_job_id as job_id,
            key        as arg_key,
            value      as arg_value 
    from json_each(arguments_in);
    
    --commit;
    
    select 'inserted job' into message_out;
    
end;$$;
""".replace("@SCHEMA_NAME", schema)

    with db_engine.connect() as conn:
        # table functions
        conn.execute(sqlalchemy.text(current_jobs_function_sql))
        conn.execute(sqlalchemy.text(insert_procedure_sql))
        conn.commit()


def _kill(db_engine: sqlalchemy.engine.base.Engine, schema: str):
    with db_engine.connect() as conn:
        conn.execute(sqlalchemy.text(f"DROP SCHEMA IF EXISTS {schema} CASCADE;"))
        conn.execute(sqlalchemy.text(f"DROP TYPE IF EXISTS jobmaster_argument CASCADE;"))
        conn.commit()





__version__ = "0.1.1"
__author__ = 'Joshua Prettyman'
__credits__ = 'Blink SEO'


from typing import Callable
from .jobs import JobMaster
from .tasks import Dependency, SameParameter
from .decorator import task
from . import utils

same = SameParameter()


def dep(_task_: str | Callable, _hours_: float = None, **kwargs):
    return Dependency(_task_, _hours_, **kwargs)



import functools

from .tasks import Dependency, Task, TaskMaster

__TASKMASTER__ = TaskMaster()


def task(_func=None, *,
         process_units: int = 1,
         write_all: str | list[str] = None,
         type_key: str = None,
         dependencies: Dependency | list[Dependency] = None
         ):

    if dependencies is None:
        dependencies = []
    if isinstance(dependencies, Dependency):
        dependencies = [dependencies]

    def decorator_task(func):
        _task: Task = Task(
            func,
            write_all=write_all,
            type_key=type_key,
            process_units=process_units,
            dependencies=dependencies
        )
        __TASKMASTER__.append(_task)

        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

    if _func is None:
        return decorator_task
    else:
        return decorator_task(_func)


import datetime
import inspect
import json
import re
import types


def wrap_string(_value):
    if isinstance(_value, str):
        return f'"{_value}"'
    return _value


def parse_params_from_doc(doc_string):
    if doc_string is None:
        return dict()
    _doc_lines = [_.strip() for _ in doc_string.split("\n")]
    params = dict()
    for _line in _doc_lines:
        if m := re.match(r":param (\w+): (.*)", _line):
            params[m.group(1)] = dict(help_str=m.group(2))
    return params


def parse_params_from_signature(signature: inspect.Signature):
    params = dict()
    for _name, _param in signature.parameters.items():

        _default = _param.default
        if _default == inspect._empty:
            _default = None
            _required = True
        else:
            _required = False

        _select_from = None
        _type = None
        if _param.annotation is None:
            _type = type(None).__name__
        elif isinstance(_param.annotation, types.UnionType | types.GenericAlias):
            _type = str(_param.annotation)
        elif isinstance(_param.annotation, type):
            _type = _param.annotation.__name__
        elif isinstance(_param.annotation, list | tuple):
            _select_from = list(_param.annotation)
            _types = set(type(x) for x in _param.annotation)
            if len(_types) == 1:
                _type = _types.pop().__name__

        params[_name] = dict(
            name=_name,
            value_type=_type,
            select_from=_select_from,
            default_value=_default,
            required=_required
        )
    return params


def parse_argument_type(arg_str: str) -> dict:
    try:
        args = json.loads(arg_str)
    except json.JSONDecodeError:
        args = arg_str

    if isinstance(args, dict):
        return args

    _out = dict()
    for _pair in args.strip('{}').strip('"')[1:-1].split(')","('):
        first_comma = _pair.index(',')
        _key, _json_value = _pair[:first_comma], _pair[first_comma + 1:]

        try:
            _v = json.loads(_json_value)
        except json.JSONDecodeError:
            _v = _json_value.strip('\\"')

        _out[_key] = _v

    return _out


class StdLogger:
    def __init__(self, name: str = "JobMaster"):
        self.name = name

    def debug(self, msg):
        print(f"{datetime.datetime.now()} | DEBUG {self.name}: {msg}")

    def info(self, msg):
        print(f"{datetime.datetime.now()} | INFO {self.name}: {msg}")

    def warning(self, msg):
        print(f"{datetime.datetime.now()} | WARNING {self.name}: {msg}")

    def error(self, msg):
        print(f"{datetime.datetime.now()} | ERROR {self.name}: {msg}")

    def critical(self, msg):
        print(f"{datetime.datetime.now()} | CRITICAL {self.name}: {msg}")


class NothingLogger:
    def debug(self, msg):
        pass

    def info(self, msg):
        pass

    def warning(self, msg):
        pass

    def error(self, msg):
        pass

    def critical(self, msg):
        pass



import inspect
import json
import re
from typing import Callable

from . import utils


class SameParameter:
    def __bool__(self):
        return True

    def __repr__(self):
        return "same"


class Dependency:
    """
    A class to represent a dependency between two tasks.

    The dependency is defined by the task, the time since the task was completed, and any specific arguments.
    """
    def __init__(self,
                 _task_: str | Callable,
                 _hours_: float = None,
                 **kwargs
                 ):
        """
        Create a new dependency object.

        :param _task_:  Either a string in the format 'type.task' or a callable function.
        :param _hours_: Hours since the task was completed. This cannot be less than 2 minutes or more than 30 days.
                        If a value less than 2 minutes is provided, it will be set to 2 minutes.
                        If a value more than 30 days is provided, it will be set to 30 days.
        :param kwargs:  Any specific arguments required for the task.
        """

        if isinstance(_task_, str):
            _keys = tuple(_task_.split('.'))
            if len(_keys) == 2:
                self.type_key, self.task_key = _keys
                self.task_func_module = None
            else:
                raise ValueError(f"Invalid task: {_task_}")
        elif isinstance(_task_, Callable):
            self.task_key = _task_.__name__
            self.type_key = None
            self.task_func_module = _task_.__module__
        else:
            raise ValueError(f"Invalid task arg: {_task_}")

        if not _hours_:
            _hours_ = 24*30
        if _hours_ > 24*30:
            _hours_ = 24*30
        if _hours_ < 2/60:
            _hours_ = 2/60
        self.time = _hours_

        self.all_arguments = kwargs
        args_same = []
        args_specified = []
        for _k, _v in kwargs.items():
            if _k == '_hours_':
                _k = 'hours'
            if isinstance(_v, SameParameter):
                args_same.append(_k)
            else:
                args_specified.append((_k, _v))

        self.args_same = args_same
        self.args_specified = args_specified

    def set_type_key(self, type_key: str):
        self.type_key = type_key

    def set_task_key(self, task_key: str):
        self.task_key = task_key

    def __repr__(self):
        return f"Dependency: {self.type_key}.{self.task_key}"


class Parameter:
    def __init__(self,
                 name: str,
                 value_type: str,
                 help_str: str,
                 select_from: list,
                 required: bool,
                 default_value,
                 write_all: bool):
        self.name = name
        self.value_type = value_type
        self.help = help_str
        self.required = required
        self.default_value = default_value
        self.select_from = select_from
        self.write_all = write_all

    def __repr__(self):
        return f"Parameter: {self.name}"

    @property
    def db_value(self):
        if self.select_from is None:
            select_from_array = "NULL"
        else:
            _select_from_json = [f"'{_v}'::json" for _v in self.select_from]
            select_from_array = f"ARRAY[{', '.join(_select_from_json)}]"

        if self.default_value is None:
            default_value = "NULL"
        else:
            default_value = f"'{self.default_value}'::json"
        return f"""
        '{self.name}', '{self.value_type}', {select_from_array}, {default_value}, '{int(self.required)}', '{int(self.write_all)}', '{self.help}'
        """.strip()


class Task:
    def __init__(self,
                 function,
                 write_all: str | list[str] = None,
                 type_key: str = None,
                 process_units: int = 1,
                 dependencies: list[Dependency] = None):

        if isinstance(write_all, str):
            write_all = [_.strip() for _ in write_all.split(',')]
        if write_all is None:
            write_all = []

        assert callable(function), f"function must be callable, not {type(function)}"
        assert isinstance(write_all, list), f"write_all must be a list, not {type(write_all)}"
        assert all(isinstance(_, str) for _ in write_all), f"write_all must be a list of strings, not {write_all}"
        assert isinstance(process_units, int), f"process_units must be an integer, not {type(process_units)}"

        if not type_key:
            type_key = function.__module__.split('.')[-1]

        if dependencies is None:
            dependencies = []

        self.function = function
        self.dependencies = dependencies

        self.type_key = type_key
        self.key = function.__name__

        self.module = function.__module__
        self.absfile = inspect.getabsfile(function)

        self.write_all = write_all
        self.process_units = process_units

        self._parameters = None
        self._help = None

    def __eq__(self, other):
        return self.module + "." + self.key == other.module + "." + other.key

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def __repr__(self):
        s = f"- Task: {self.type_key}.{self.key}"
        if self.process_units:
            s += f"\n  Process units: {self.process_units}"
        if self.parameters:
            s += f"\n  Parameters:"
            for i, _param in enumerate(self.parameters):
                s += f"\n\t  {i+1}. {_param.name}: {_param.value_type} ({_param.help})"
        if self.dependencies:
            s += f"\n  Dependencies:"
            for i, _dep in enumerate(self.dependencies):
                _args = ", ".join(
                    [
                        f"{_arg}={_val}"
                        if not isinstance(_val, str)
                        else f"{_arg}=\"{_val}\""
                        for _arg, _val in _dep.all_arguments.items()
                    ]
                )
                s += f"\n\t  {i+1}. {_dep.type_key}.{_dep.task_key}("
                s += _args
                s += ")"
        return s

    def _parse_parameters(self) -> list[Parameter]:
        _docstring = inspect.getdoc(self.function)
        _signature = inspect.signature(self.function)

        params_from_doc = utils.parse_params_from_doc(_docstring)
        params_from_sig = utils.parse_params_from_signature(_signature)
        parameters = []
        for _arg_name in params_from_sig.keys():
            if _arg_name in ('self', 'cls', 'kwargs', 'args'):
                continue

            _param = params_from_sig[_arg_name]
            for _k, _v in params_from_doc.get(_arg_name, dict()).items():
                if _v:
                    if not _param.get(_k):
                        _param[_k] = _v

            if _param.get('help_str') is None:
                _param['help_str'] = "No documentation for parameter"

            if _arg_name in self.write_all and isinstance(_param.get('select_from'), list):
                _param['write_all'] = True
            else:
                _param['write_all'] = False

            _param_object = Parameter(**_param)

            parameters.append(_param_object)

        return parameters

    def _parse_help(self) -> str:
        doc = inspect.getdoc(self.function)
        if not doc:
            return "No docstring provided"
        else:
            return re.sub(r"\s+", " ", re.split(r"\n\s*\n", doc)[0]).strip()

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = self._parse_parameters()
        return self._parameters

    @property
    def parameter_keys(self):
        return [_.name for _ in self.parameters]

    @property
    def parameter_dict(self):
        return {_.name: _ for _ in self.parameters}

    @property
    def help(self):
        if self._help is None:
            self._help = self._parse_help()
        return self._help

    @property
    def db_dependency_rows(self) -> tuple[list[str], list[str], list[str]]:
        dep_rows = [
            f"'{self.type_key}', '{self.key}', '{_dep.type_key}', '{_dep.task_key}', {_dep.time}"
            for _dep in self.dependencies
        ]

        dep_args_same_rows = [
            f"'{self.type_key}', '{self.key}', '{_dep.type_key}', '{_dep.task_key}', '{_arg}'"
            for _dep in self.dependencies
            for _arg in _dep.args_same
        ]

        dep_args_specified_rows = [
            f"'{self.type_key}', '{self.key}', '{_dep.type_key}', '{_dep.task_key}', '{_arg[0]}', '{json.dumps(_arg[1])}'"
            for _dep in self.dependencies
            for _arg in _dep.args_specified
        ]

        return dep_rows, dep_args_same_rows, dep_args_specified_rows

    @property
    def db_parameter_rows(self) -> list[str]:
        return [f"'{self.type_key}', '{self.key}', " + _p.db_value for _p in self.parameters]

    @property
    def db_row(self) -> str:
        if not self.write_all:
            _write_all = "NULL"
        else:
            _write_all = f"ARRAY['{', '.join(self.write_all)}']"

        return f"'{self.type_key}', '{self.key}', '{self.help}', {self.process_units}, {_write_all}"


class TaskType:
    def __init__(self, key: str, help_str: str = None):
        self._key: str = key
        self._help_str: str = help_str
        self._tasks: dict[str, Task] = dict()

    @property
    def key(self):
        return self._key

    @property
    def tasks(self):
        return self._tasks

    def __getitem__(self, key: str):
        return self._tasks.get(key)

    def append(self, task: Task):
        self._tasks[task.key] = task


class TaskMaster:
    def __init__(self):
        self._tasks_types: dict[str: TaskType] = dict()
        self._dependencies_validated: bool = False

    @property
    def task_types(self):
        return self._tasks_types

    @property
    def dependencies_validated(self) -> bool:
        return self._dependencies_validated

    def __getitem__(self, key: str | tuple[str, str]):
        if isinstance(key, str):
            keys = tuple(key.split('.'))
        elif isinstance(key, tuple):
            keys = key
        else:
            raise ValueError(f"bad task key {key}")

        if len(keys) == 2:
            _tt = self._tasks_types.get(keys[0])
            if _tt is None:
                return None
            else:
                return _tt[keys[1]]
        elif len(keys) == 1:
            return self._tasks_types.get(key)
        else:
            raise ValueError(f"bad task key {key}")

    @property
    def _task_lookup(self):
        _lookup = dict()
        for _type_key, _task_type in self._tasks_types.items():
            for _task_key, _task in _task_type.tasks.items():
                _lookup[_task.module+'.'+_task.key] = (_type_key, _task_key)
        return _lookup

    def all_tasks(self):
        for _type_key, _task_type in self._tasks_types.items():
            for _task_key, _task in _task_type.tasks.items():
                yield _type_key, _task_key, _task

    def print_tasks(self):
        for _type_key, _task_key, _task in self.all_tasks():
            print(_task)
            print('')

    def append(self, task: Task):
        if task.type_key not in self._tasks_types.keys():
            self._tasks_types[task.type_key] = TaskType(task.type_key)
        self._tasks_types[task.type_key].append(task)

    def validate_dependencies(self):
        for _type_key, _task_type in self._tasks_types.items():
            for _task_key, _task in _task_type.tasks.items():
                for _dep in _task.dependencies:
                    if _dep.type_key is None:
                        __type, __task = self._task_lookup.get(_dep.task_func_module+'.'+_dep.task_key, (None, None))
                        _dep.set_type_key(__type)
                    if not self.__getitem__((_dep.type_key, _dep.task_key)):
                        raise ValueError(f"Dependency {_dep.type_key}.{_dep.task_key} for "
                                         f"task {_type_key}.{_task_key} not recognised")

        self._dependencies_validated = True



from setuptools import setup

# python setup.py check
# python setup.py sdist
# python setup.py bdist_wheel --universal
# twine upload dist/*

_desc = """JobMaster allows a user to quickly and simply deploy and manage a collection of tasks."""

setup(
    name='jobmaster',
    version='0.1.1',
    description=_desc,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/DrPrettyman/jobmaster',
    author='DrPrettyman',
    author_email='joshua@blinkseo.co.uk',
    license='MIT',
    packages=['jobmaster'],
    install_requires=[
        'pg8000',
        'sqlalchemy'
    ],
    classifiers=[
        "Intended Audience :: Developers",
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
)



import datetime
import json
import os
import uuid
import sqlalchemy
from uuid import uuid4
from typing import Callable
from itertools import product
from sqlalchemy.sql.elements import TextClause

from . import utils
from . import tasks
from .decorator import __TASKMASTER__
from .deploy import deploy


__SYSTEM_NODE_NAME__ = os.uname().nodename.split('.')[0]
__SYSTEM_PID__ = os.getpid()
_status_lookup = {
    0: 'local',
    1: 'waiting',
    2: 'running',
    3: 'complete',
    4: 'failed',
    5: 'cancelled'
}


class Job:
    def __init__(self,
                 cmd_id: str,
                 db_engine: sqlalchemy.engine.base.Engine,
                 db_schema: str,
                 task: tasks.Task,
                 logger: Callable,
                 priority: int = 0,
                 status: int = 0,
                 arguments: dict | str = None,
                 job_id: str = None,
                 created_at: datetime.datetime = None,
                 collect_id: str = None,
                 message: str = None,
                 fill_args: bool = False):

        if job_id is None:
            job_id = uuid4().hex
        self.job_id = job_id

        if arguments is None:
            arguments = dict()
        if isinstance(arguments, str):
            arguments = utils.parse_argument_type(arguments)
        if not isinstance(arguments, dict):
            raise ValueError(f"Invalid arguments: {arguments}")
        self.arguments = arguments

        self.logger = logger

        self._engine = db_engine
        self._schema = db_schema

        self._cmd_id = cmd_id

        self.task = task
        self.priority = priority

        self._status = status
        if created_at is None:
            created_at = datetime.datetime.utcnow()
        self.created_at = created_at

        self.collect_id = collect_id
        self.message = message

        # fill in required arguments
        if fill_args:
            for _param in self.task.parameters:
                if not _param.required:
                    continue  # If the argument is not required, we don't need to fill it
                if _param.name not in self.arguments.keys():
                    if _param.default_value is not None:
                        self.arguments[_param.name] = _param.default_value
                        self.logger.debug(f"Required argument `{_param.name}` not provided. "
                                          f"Assigned default value `{_param.default_value}` ({type(_param.default_value)}).")
                    elif _param.value_type == 'NoneType':
                        self.arguments[_param.name] = None
                        self.logger.debug(f"Required argument `{_param.name}` not provided. "
                                          f"Assigned value `None`.")
                    elif _param.select_from:
                        self.arguments[_param.name] = _param.select_from[0]
                        self.logger.debug(f"Required argument `{_param.name}` not provided. "
                                          f"Assigned value `{_param.select_from[0]}` ({type(_param.select_from[0])}).")
                    else:
                        self.logger.error(f"Argument `{_param.name}` is required but not provided")

        self.logger.info(f"Initialised local Job instance with ID: {self.job_id}")

    def __bool__(self):
        return True

    def __repr__(self):
        args = ", ".join(
            [f"{_arg}={_val}"
             if not isinstance(_val, str)
             else f"{_arg}=\"{_val}\"" for _arg, _val
             in self.arguments.items()]
        )
        lines = [
            f"Job ID:    {self.job_id}"
            f"\nSignature: {self.task.type_key}.{self.task.key}({args})"
            f"\nStatus:    {self.status_str}"
        ]
        if self.message:
            lines.append(f"\nMessage:   {self.message}")

        return "".join(lines)

    def __str__(self):
        return self.__repr__()

    @property
    def active(self):
        return self._status in (0, 1, 2)

    def set_message(self, message: str, _message_level: str = 'debug', _status: int = None):
        if _status is None:
            _status_str = self.status_str
        else:
            _status_str = _status_lookup.get(_status)

        message = f"[{_status_str}] {message}"

        if _message_level == 'debug':
            self.logger.debug(f"Job {self.job_id}: {message}")
        elif _message_level == 'info':
            self.logger.info(f"Job {self.job_id}: {message}")
        elif _message_level == 'warning':
            self.logger.warning(f"Job {self.job_id}: {message}")
        elif _message_level == 'error':
            self.logger.error(f"Job {self.job_id}: {message}")
        elif _message_level == 'critical':
            self.logger.critical(f"Job {self.job_id}: {message}")

        self.message = message

    @property
    def status(self):
        return self._status

    @property
    def status_str(self):
        return _status_lookup[self._status]

    def arguments_for_spawned_jobs(self) -> list[dict]:
        options = dict()

        parameters_with_all_value = [
            _key
            for _key in self.task.write_all
            if self.arguments.get(_key) == 'ALL'
        ]

        for _param_key in parameters_with_all_value:
            _parameter = self.task.parameter_dict.get(_param_key)
            if _parameter is None:
                self.update(status=4, message=f"Parameter {_param_key} not recognised")
                self.logger.error(f"Parameter {_param_key} not recognised")
                return []
            _arg_options = _parameter.select_from
            if len(_arg_options) == 0:
                self.update(status=4, message=f"Parameter {_param_key} has no options")
                self.logger.error(f"Parameter {_param_key} has no options")
                return []
            options[_param_key] = _arg_options

        if len(options) == 0:
            return []

        # create jobs with all combinations of options
        option_tuples = []
        for k, v in options.items():
            option_tuples.append([(k, sv) for sv in v])
        new_job_options = [{k: v for k, v in _tup} for _tup in [_p for _p in product(*option_tuples)]]
        new_args = []
        for _opt in new_job_options:
            _args = self.arguments.copy()
            _args.update(_opt)
            new_args.append(_args)

        return new_args

    @property
    def _db_rows(self) -> tuple[str, list[str]]:
        if self.collect_id is None:
            _collect_id = "NULL"
        else:
            _collect_id = f"'{self.collect_id}'"

        if self.message is None:
            _message = "NULL"
        else:
            _m = self.message.replace("'", "`")
            _message = f"'{_m}'"

        queue_row = f"('{self.job_id}', {_collect_id}, {self._status}, '{__SYSTEM_NODE_NAME__}', {__SYSTEM_PID__}, '{self._cmd_id}', '{datetime.datetime.utcnow()}', {self.priority}, '{self.task.type_key}', '{self.task.key}', {_message})"
        arg_rows = [f"('{self.job_id}', '{_k}', '{json.dumps(_v)}'::json)" for _k, _v in self.arguments.items()]

        return queue_row, arg_rows

    @property
    def _dependencies_sql_table_simple(self) -> str:
        selects = [
            f"SELECT '{_dep.type_key}' AS dependency_type_key, '{_dep.task_key}' AS dependency_task_key "
            for _dep in self.task.dependencies
        ]
        return " UNION ALL ".join(selects)

    @property
    def _dependencies_sql_table(self) -> str:
        selects = []
        for _dep in self.task.dependencies:
            args = [(_k, _v) for _k, _v in _dep.args_specified]
            for _k in _dep.args_same:
                if _k in self.arguments.keys():
                    args.append((_k, self.arguments[_k]))
                else:
                    raise ValueError(f"Dependency argument {_k} not found in job arguments")

            arg_rows = [f"ROW('{_k}', '{json.dumps(_v)}'::json)::jobmaster_argument" for _k, _v in args]
            arg_array = f"ARRAY[{', '.join(arg_rows)}]"

            sql = f"""
                    SELECT  '{self.task.type_key}' AS type_key,
                            '{self.task.key}' AS task_key,
                            '{_dep.type_key}' AS dependency_type_key,
                            '{_dep.task_key}' AS dependency_task_key,
                            {_dep.time} AS dependency_time,
                            {arg_array} AS arguments
                    """
            selects.append(sql)
        return " UNION ALL ".join(selects)

    def update(self,
               status: int,
               message: str = None,
               _execute: bool = True,
               _message_level: str = 'debug') -> TextClause | None:
        """
        Update the status of the job and write to the database.
        If the status is 0 (local), the arguments will also be written to the database,

        :param status:
        :param message:
        :param _execute:
        :param _message_level:
        :return:
        """

        if status not in _status_lookup.keys():
            raise ValueError(f"Invalid status: {status}")

        if message:
            self.set_message(message, _message_level=_message_level, _status=status)

        if status == self._status:
            return

        if self._status == 0:
            write_args = True
        else:
            write_args = False

        self._status = status
        self.logger.debug(f"Job {self.job_id} status set: {self.status_str}")

        queue_row, arg_rows = self._db_rows

        if write_args:
            sql = sqlalchemy.text(
                f"INSERT INTO {self._schema}.jobs "
                f"VALUES {queue_row}; "
                f"INSERT INTO {self._schema}.arguments "
                f"VALUES {', '.join(arg_rows)}; "
            )
        else:
            sql = sqlalchemy.text(
                f"INSERT INTO {self._schema}.jobs "
                f"VALUES {queue_row}; "
            )

        if _execute:
            with self._engine.connect() as conn:
                conn.execute(sql)
                conn.commit()
            return
        else:
            return sql

    def dependencies_in_queue(self) -> list[dict]:
        with self._engine.connect() as conn:
            _result = conn.execute(
                sqlalchemy.text(
                    f"""
                    WITH jobs_table AS (
                        SELECT *
                        FROM {self._schema}.current_jobs()
                        WHERE status_out IN (1,2,3)
                    ), deps_table AS (
                        {self._dependencies_sql_table_simple}
                    )

                    SELECT  dt.dependency_type_key as type_key,
                            dt.dependency_task_key as task_key,
                            jt.job_id_out as job_id,
                            jt.status_out as status,
                            jt.updated_at_out as updated_at,
                            jt.arguments_out as arguments
                    FROM deps_table dt
                    INNER JOIN jobs_table jt
                    ON dt.dependency_type_key = jt.type_key_out
                    AND dt.dependency_task_key = jt.task_key_out
                    """
                )
            )
            conn.commit()
        rows = _result.all()
        self.logger.debug(f"Dependencies in queue: {len(rows)}")
        if len(rows) > 0:
            results = [
                dict(
                    job_id=row.job_id,
                    status=row.status,
                    updated_at=row.updated_at,
                    type_key=row.type_key,
                    task_key=row.task_key,
                    arguments=utils.parse_argument_type(row.arguments)
                )
                for row in rows
            ]
        else:
            results = []

        return results

    def dependencies_specific(self) -> list[dict]:
        _deps = []
        for i, _dep in enumerate(self.task.dependencies):
            _d = dict(
                type_key=_dep.type_key,
                task_key=_dep.task_key,
                time=_dep.time
            )
            _args = dict()
            for _k in _dep.args_same:
                _args[_k] = self.arguments[_k]
            for _k, _v in _dep.args_specified:
                _args[_k] = _v
            _d['arguments'] = _args
            _deps.append(_d)
        return _deps

    def required_dependency_jobs(self):
        required = []
        for _dep in self.dependencies_specific():
            _dep_is_required = True
            for _qj in self.dependencies_in_queue():
                if _dep['type_key'] == _qj['type_key'] and _dep['task_key'] == _qj['task_key'] and all(
                        (k, v) in _qj['arguments'].items() for k, v in _dep['arguments'].items()):
                    if _qj['status'] in (1, 2):
                        # If the job is already in the queue (waiting/running), it is not required
                        _dep_is_required = False
                    elif _qj['status'] == 3:
                        # If the job is complete, check if it was completed within the required time
                        time_since_complete = (datetime.datetime.utcnow() - _qj['updated_at']).total_seconds() / 3600
                        if time_since_complete < _dep['time']:
                            # If the job was completed within the required time, it is not required
                            _dep_is_required = False
            if _dep_is_required:
                required.append(_dep)
        self.logger.debug(f"Required dependencies: {len(required)}")
        return required

    def safe_execute(self):
        self.logger.debug(f"Executing job {self.job_id}")
        _success = False
        try:
            result = self.task.function(**self.arguments)
            self.update(status=3, message=f"Executed at {datetime.datetime.utcnow()} with result: {str(result)}")
            _success = True
        except Exception as _error:
            self.update(status=4, message=f"Error at {datetime.datetime.utcnow()}: {_error}", _message_level='error')
        return _success


class JobMaster:
    def __init__(self,
                 db_engine: sqlalchemy.engine.base.Engine,
                 schema: str = 'jobmaster',
                 cmd_id: str = None,
                 system_process_units: int = None,
                 logger=None,
                 _validate_dependencies: bool = False):

        if logger is None:
            self.logger = utils.NothingLogger()
        elif logger == print:
            self.logger = utils.StdLogger()
        elif all(hasattr(logger, _attr) for _attr in ('debug', 'info', 'warning', 'error', 'critical')):
            self.logger = logger
        else:
            raise ValueError("Invalid logger")

        self._engine = db_engine
        self._schema = schema
        self._cmd_id = cmd_id or datetime.datetime.utcnow().isoformat()

        if system_process_units is None:
            _ = os.environ.get('JOBMASTER_SYSTEM_PROCESS_UNITS', '10_000')
            try:
                system_process_units = int(_)
            except ValueError:
                system_process_units = 10_000

        assert isinstance(system_process_units, int), "system_process_units must be an integer"

        self._system_process_units = system_process_units

        if _validate_dependencies:
            self._validate_dependencies()

    @property
    def cmd_id(self):
        return self._cmd_id

    @property
    def system_process_units(self):
        return self._system_process_units

    def __getitem__(self, key: str | tuple[str, str]):
        return __TASKMASTER__[key]

    @property
    def dependencies_validated(self) -> bool:
        return __TASKMASTER__.dependencies_validated

    @staticmethod
    def _validate_dependencies():
        __TASKMASTER__.validate_dependencies()

    @staticmethod
    def print_tasks():
        __TASKMASTER__.print_tasks()

    def write_tasks(self):
        """
        Writes all tasks to the database. This will overwrite any existing tasks.
        Task data is written to the tables `tasks` and `parameters` in the schema specified at initialisation.
        """

        if not self.dependencies_validated:
            self._validate_dependencies()

        statements = []

        _sql = sqlalchemy.text(
            f"""
            TRUNCATE {self._schema}.tasks, 
                     {self._schema}.parameters, 
                     {self._schema}.dependencies,
                     {self._schema}.dependency_args_specified,
                     {self._schema}.dependency_args_same;
            """
        )

        statements.append(_sql)

        task_rows = ["(" + _task.db_row + ")" for _type_key, _task_key, _task in __TASKMASTER__.all_tasks()]
        if len(task_rows) > 0:
            _sql = sqlalchemy.text(f"INSERT INTO {self._schema}.tasks \nVALUES \n" + ", \n".join(task_rows) + ";")
            statements.append(_sql)

        raw_param_rows = [_task.db_parameter_rows for _type_key, _task_key, _task in __TASKMASTER__.all_tasks()]
        param_rows = []
        for _l in raw_param_rows:
            param_rows.extend(["(" + _r + ")" for _r in _l])

        if len(param_rows) > 0:
            _sql = sqlalchemy.text(f"INSERT INTO {self._schema}.parameters \nVALUES \n" + ", \n".join(param_rows) + ";")
            statements.append(_sql)

        raw_dep_rows = [_task.db_dependency_rows for _type_key, _task_key, _task in __TASKMASTER__.all_tasks()]
        dep_rows = []
        dep_arg_same_rows = []
        dep_arg_specified_rows = []
        for _tup in raw_dep_rows:
            dep_rows.extend(["(" + _r + ")" for _r in _tup[0]])
            dep_arg_same_rows.extend(["(" + _r + ")" for _r in _tup[1]])
            dep_arg_specified_rows.extend(["(" + _r + ")" for _r in _tup[2]])

        if len(dep_rows) > 0:
            _sql = sqlalchemy.text(
                f"INSERT INTO {self._schema}.dependencies \nVALUES \n" + ", \n".join(dep_rows) + ";"
            )
            statements.append(_sql)
        if len(dep_arg_same_rows) > 0:
            _sql = sqlalchemy.text(
                f"INSERT INTO {self._schema}.dependency_args_same \nVALUES \n" + ", \n".join(dep_arg_same_rows) + ";"
            )
            statements.append(_sql)
        if len(dep_arg_specified_rows) > 0:
            _sql = sqlalchemy.text(
                f"INSERT INTO {self._schema}.dependency_args_specified \nVALUES \n"
                + ", \n".join(dep_arg_specified_rows) + ";"
            )
            statements.append(_sql)

        with self._engine.connect() as conn:
            for _sql in statements:
                conn.execute(_sql)
            conn.commit()

    def deploy(self, reset: bool = False):

        if not self.dependencies_validated:
            self._validate_dependencies()

        deploy(self._engine, self._schema, reset=reset)
        if reset:
            self.logger.info("Reset JobMaster database")
        self.logger.info("Deployed JobMaster database")
        self.write_tasks()

    def job(self,
            type_key: str,
            task_key: str,
            priority: int = 0,
            status: int = 0,
            arguments: dict = None,
            job_id: str = None,
            created_at: datetime.datetime = None,
            collect_id: str = None,
            message: str = None,
            fill_args: bool = False
            ) -> Job:

        _task = __TASKMASTER__[(type_key, task_key)]
        if not _task:
            raise ValueError(f"Task {type_key}.{task_key} not recognised")

        return Job(
            cmd_id=self._cmd_id,
            db_engine=self._engine,
            db_schema=self._schema,
            task=_task,
            priority=priority,
            status=status,
            arguments=arguments,
            job_id=job_id,
            created_at=created_at,
            collect_id=collect_id,
            message=message,
            logger=self.logger,
            fill_args=fill_args
        )

    def jobs_in_queue(self) -> list[Job]:
        """
        Simply returns all jobs in the queue, ordered by "updated_at" time.

        :return:
        """
        with self._engine.connect() as conn:
            _result = conn.execute(
                sqlalchemy.text(
                    f"SELECT * "
                    f"FROM {self._schema}.current_jobs() "
                    f"ORDER BY updated_at_out DESC "
                )
            )
            conn.commit()

        results = _result.all()

        if len(results) == 0:
            self.logger.info("No jobs in queue")
            return []

        self.logger.info(f"{len(results)} jobs in queue")
        _out = [
            self.job(
                job_id=_r.job_id_out,
                type_key=_r.type_key_out,
                task_key=_r.task_key_out,
                status=_r.status_out,
                priority=_r.priority_out,
                created_at=_r.created_at_out,
                arguments=_r.arguments_out
            )
            for _r in results
        ]

        return _out

    def available_process_units(self) -> int:
        with self._engine.connect() as conn:
            _processes_result = conn.execute(
                sqlalchemy.text(
                    f"""
                    WITH running_jobs AS (
                      SELECT type_key_out, task_key_out 
                      FROM {self._schema}.current_jobs() 
                      WHERE status_out = 2 
                      AND system_node_name_out = '{__SYSTEM_NODE_NAME__}' 
                    ) 
                    SELECT SUM(tsk.process_units) AS total_process_units 
                    FROM running_jobs 
                    LEFT JOIN {self._schema}.tasks tsk 
                    ON running_jobs.type_key_out = tsk.type_key 
                    AND running_jobs.task_key_out = tsk.task_key; 
                """
                )
            )
            conn.commit()

        _ = _processes_result.all()
        _process_units_in_use = None
        if len(_) > 0:
            _process_units_in_use = _[0].total_process_units
        if _process_units_in_use is None:
            _process_units_in_use = 0
        return self._system_process_units - _process_units_in_use

    def queue_pop(self) -> Job | None:
        """
        Pop the top job from the queue and return it as a Job object.
        """
        if not self.dependencies_validated:
            self._validate_dependencies()

        # Taking out the lines below because the check on process units
        # is now within the main sql query.

        # available_process_units = self.available_process_units()
        # if available_process_units <= 0:
        #     self.logger.info("No available process units")
        #     return None

        collect_id: str = str(uuid.uuid4())
        with self._engine.connect() as conn:
            # first we lock the collect_id table, this prevents other instances trying to run the queue_pop() method
            conn.execute(
                sqlalchemy.text(
                    f"LOCK TABLE {self._schema}.collect_ids IN ACCESS EXCLUSIVE MODE; "
                    f"INSERT INTO {self._schema}.collect_ids (collect_id) VALUES ('{collect_id}');"
                )
            )
            _result = conn.execute(
                sqlalchemy.text(
                    f"""
                    WITH waiting_and_running_jobs AS (
                        SELECT    job_id_out AS job_id,
                                  type_key_out AS type_key, 
                                  task_key_out AS task_key, 
                                  status_out AS status, 
                                  priority_out AS priority, 
                                  created_at_out AS created_at, 
                                  arguments_out AS arguments,
                                  system_node_name_out AS system_node_name 
                        FROM {self._schema}.current_jobs() 
                        WHERE status_out in (1, 2) 
                    ), waiting_jobs AS (
                        SELECT   job_id, type_key, task_key, status, priority, created_at, arguments
                        FROM     waiting_and_running_jobs
                        WHERE    status = 1
                    ), jobs_running_on_system AS (
                        SELECT   type_key, task_key
                        FROM     waiting_and_running_jobs
                        WHERE    status = 2
                        AND      system_node_name = '{__SYSTEM_NODE_NAME__}' 
                    ), p_units AS (
                        SELECT  SUM(tsk.process_units) AS total_process_units 
                        FROM    jobs_running_on_system jsys
                        LEFT JOIN {self._schema}.tasks tsk 
                        ON  jsys.type_key = tsk.type_key 
                        AND jsys.task_key = tsk.task_key
                    )
                    SELECT  wj.job_id,
                            wj.type_key, 
                            wj.task_key, 
                            wj.status, 
                            wj.priority, 
                            wj.created_at, 
                            wj.arguments,
                            tsk.process_units 
                    FROM waiting_jobs wj 
                    LEFT JOIN {self._schema}.tasks tsk 
                    ON  wj.type_key = tsk.type_key 
                    AND wj.task_key = tsk.task_key 
                    WHERE tsk.process_units <= {self._system_process_units} - COALESCE((SELECT total_process_units FROM p_units), 0)
                    ORDER BY priority DESC, created_at ASC 
                    LIMIT 1; 
                    """
                )
            )
            # Create a new job from the result
            results = _result.all()
            if len(results) == 0:
                self.logger.info("Queue is empty. No jobs to run.")
                new_job = None
            else:
                top_job_out = results[0]
                self.logger.info(f"Found a job to run: {top_job_out.job_id}")
                new_job = self.job(
                    job_id=top_job_out.job_id,
                    collect_id=collect_id,
                    type_key=top_job_out.type_key,
                    task_key=top_job_out.task_key,
                    status=top_job_out.status,
                    priority=top_job_out.priority,
                    created_at=top_job_out.created_at,
                    arguments=top_job_out.arguments
                )
                # Execute the update status for this job within this same transaction,
                # while the collect_ids table is still locked.
                sql = new_job.update(status=2, message="Popped from queue", _execute=False)
                conn.execute(sql)
            conn.commit()
        return new_job

    def update(self, jobs: Job | list[Job], status: int, message: str = None, _message_level: str = 'debug'):
        """
        Performs the same function as Job.update(), but for potentially multiple jobs.
        If the status of a job is 0 (local), the arguments will be written to the database,
        otherwise only the job status will be updated.

        :param jobs: a Job or list of Jobs
        :param status: The status to update to - this will be the same for all if updating multiple jobs
        :param message: A message to write to the database - this will be the same for all if updating multiple jobs
        :param _message_level: The logging level to use for the message
        :return:
        """

        if isinstance(jobs, Job):
            jobs = [jobs]

        queue_rows = []
        arg_rows = []
        for _j in jobs:
            _old_status = _j.status
            _j.update(status=status, message=message, _execute=False, _message_level=_message_level)
            _q, _a = _j._db_rows
            queue_rows.append(_q)
            if _old_status == 0:
                arg_rows.extend(_a)

        with self._engine.connect() as conn:
            _queue_values = ', \n'.join(queue_rows)
            conn.execute(sqlalchemy.text(f"INSERT INTO {self._schema}.jobs \nVALUES \n{_queue_values};"))
            if arg_rows:
                _arg_values = ', \n'.join(arg_rows)
                conn.execute(sqlalchemy.text(f"INSERT INTO {self._schema}.arguments \nVALUES \n{_arg_values};"))
            conn.commit()

    def execute(self, job: Job) -> Job:
        """
        For a given job, check if it can be executed.
        If the job contains any 'ALL' arguments, we create a bunch of new jobs with all combinations of options.
        If the job has dependencies, we create new jobs to satisfy them.
        If neither of the above, we execute the job.

        :param job: a Job object
        :return: the updated Job object
        """

        # update the job status to "running" if not already
        job.update(2)

        # first check "write all" parameters
        new_job_args = job.arguments_for_spawned_jobs()
        if len(new_job_args) > 0:
            # create jobs with all combinations of options
            new_jobs = [
                self.job(
                    type_key=job.task.type_key,
                    task_key=job.task.key,
                    priority=job.priority,
                    arguments=_args
                )
                for _args in new_job_args
            ]
            # write jobs to database
            self.update(new_jobs, status=1, message=f"Batch-spawned from job {job.job_id}")
            # Update this job to "completed"
            job.update(status=3, message=f"Created {len(new_jobs)} new jobs with all options")

        if job.status != 2:
            return job

        # then check dependencies
        new_job_specs = job.required_dependency_jobs()
        if len(new_job_specs) > 0:
            self.logger.debug(f"Creating {len(new_job_specs)} new jobs to satisfy dependencies")
            # create jobs to satisfy dependencies with higher priority
            new_jobs = [
                self.job(
                    type_key=_nj['type_key'],
                    task_key=_nj['task_key'],
                    priority=job.priority + 1,
                    arguments=_nj['arguments'],
                    fill_args=True
                )
                for _nj in new_job_specs
            ]
            # write jobs to database
            self.update(new_jobs, status=1, message=f"Created to satisfy dependencies of job {job.job_id}")
            # Update this job back to "waiting"
            job.update(status=1, message=f"Back in queue. Created {len(new_jobs)} new jobs to satisfy dependencies")

        if job.status != 2:
            self.logger.debug(f"Job {job.job_id} not ready to execute")
            return job

        # execute job
        job.safe_execute()
        return job

    def run(self) -> list[Job]:
        """
        Keep popping jobs from the queue and executing them until
        the queue is empty.

        :return:
        """

        if not self.dependencies_validated:
            self._validate_dependencies()
        jobs = []
        while True:
            job = self.queue_pop()
            if job is None:
                break
            _j = self.execute(job)
            jobs.append(_j)
        self.logger.info(f"Ran {len(jobs)} jobs.")
        return jobs

    def __call__(self) -> list[Job]:
        return self.run()



